/******************************************************************************
 * Multiverse 2 Copyright (c) the Multiverse Team 2011.                       *
 * Multiverse 2 is licensed under the BSD License.                            *
 * For more information please check the README.md file included              *
 * with this project.                                                         *
 ******************************************************************************/

package com.onarandombox.MultiverseCore.commands;

import com.dumptruckman.minecraft.util.Logging;
import com.onarandombox.MultiverseCore.MultiverseCore;
import com.onarandombox.MultiverseCore.event.MVVersionEvent;
import com.onarandombox.MultiverseCore.utils.webpaste.PasteFailedException;
import com.onarandombox.MultiverseCore.utils.webpaste.PasteService;
import com.onarandombox.MultiverseCore.utils.webpaste.PasteServiceFactory;
import com.onarandombox.MultiverseCore.utils.webpaste.PasteServiceType;
import com.onarandombox.MultiverseCore.utils.webpaste.URLShortener;
import com.onarandombox.MultiverseCore.utils.webpaste.URLShortenerFactory;
import com.onarandombox.MultiverseCore.utils.webpaste.URLShortenerType;
import com.pneumaticraft.commandhandler.CommandHandler;
import org.apache.commons.lang.StringUtils;
import org.bukkit.ChatColor;
import org.bukkit.command.CommandSender;
import org.bukkit.command.ConsoleCommandSender;
import org.bukkit.entity.Player;
import org.bukkit.permissions.PermissionDefault;
import org.bukkit.scheduler.BukkitRunnable;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Dumps version info to the console.
 */
public class VersionCommand extends MultiverseCommand {
    private static final URLShortener SHORTENER = URLShortenerFactory.getService(URLShortenerType.BITLY);

    public VersionCommand(MultiverseCore plugin) {
        super(plugin);
        this.setName("Multiverse Version");
        this.setCommandUsage("/mv version " + ChatColor.GOLD + "[-b|-h|-p] [--include-plugin-list]");
        this.setArgRange(0, 2);
        this.addKey("mv version");
        this.addKey("mvver");
        this.addKey("mvv");
        this.addKey("mvversion");
        this.setPermission("multiverse.core.version",
                "Dumps version info to the console, optionally to pastebin.com with -b, to hastebin.com using -h, or to paste.gg with -p.", PermissionDefault.TRUE);
    }

    private String getLegacyString() {
        return "[Multiverse-Core] Multiverse-Core Version: " + this.plugin.getDescription().getVersion() + System.lineSeparator()
                + "[Multiverse-Core] Bukkit Version: " + this.plugin.getServer().getVersion() + System.lineSeparator()
                + "[Multiverse-Core] Loaded Worlds: " + this.plugin.getMVWorldManager().getMVWorlds() + System.lineSeparator()
                + "[Multiverse-Core] Multiverse Plugins Loaded: " + this.plugin.getPluginCount() + System.lineSeparator()
                +"[Multiverse-Core] Economy being used: " + plugin.getEconomist().getEconomyName() + System.lineSeparator()
                + "[Multiverse-Core] Permissions Plugin: " + this.plugin.getMVPerms().getType() + System.lineSeparator()
                + "[Multiverse-Core] Dumping Config Values: (version " + this.plugin.getMVConfig().getVersion() + ")" + System.lineSeparator()
                + "[Multiverse-Core]   enforceaccess: " + plugin.getMVConfig().getEnforceAccess() + System.lineSeparator()
                + "[Multiverse-Core]   prefixchat: " + plugin.getMVConfig().getPrefixChat() + System.lineSeparator()
                + "[Multiverse-Core]   prefixchatformat: " + plugin.getMVConfig().getPrefixChatFormat() + System.lineSeparator()
                + "[Multiverse-Core]   useasyncchat: " + plugin.getMVConfig().getUseAsyncChat() + System.lineSeparator()
                + "[Multiverse-Core]   teleportintercept: " + plugin.getMVConfig().getTeleportIntercept() + System.lineSeparator()
                + "[Multiverse-Core]   firstspawnoverride: " + plugin.getMVConfig().getFirstSpawnOverride() + System.lineSeparator()
                + "[Multiverse-Core]   displaypermerrors: " + plugin.getMVConfig().getDisplayPermErrors() + System.lineSeparator()
                + "[Multiverse-Core]   globaldebug: " + plugin.getMVConfig().getGlobalDebug() + System.lineSeparator()
                + "[Multiverse-Core]   silentstart: " + plugin.getMVConfig().getSilentStart() + System.lineSeparator()
                + "[Multiverse-Core]   messagecooldown: " + plugin.getMessaging().getCooldown() + System.lineSeparator()
                + "[Multiverse-Core]   version: " + plugin.getMVConfig().getVersion() + System.lineSeparator()
                + "[Multiverse-Core]   firstspawnworld: " + plugin.getMVConfig().getFirstSpawnWorld() + System.lineSeparator()
                + "[Multiverse-Core]   teleportcooldown: " + plugin.getMVConfig().getTeleportCooldown() + System.lineSeparator()
                + "[Multiverse-Core]   defaultportalsearch: " + plugin.getMVConfig().isUsingDefaultPortalSearch() + System.lineSeparator()
                + "[Multiverse-Core]   portalsearchradius: " + plugin.getMVConfig().getPortalSearchRadius() + System.lineSeparator()
                + "[Multiverse-Core]   autopurge: " + plugin.getMVConfig().isAutoPurgeEnabled() + System.lineSeparator()
                + "[Multiverse-Core] Special Code: FRN002" + System.lineSeparator();
    }

    private String getMarkdownString() {
        return "# Multiverse-Core" + System.lineSeparator()
                + "## Overview" + System.lineSeparator()
                + "| Name | Value |" + System.lineSeparator()
                + "| --- | --- |" + System.lineSeparator()
                + "| Multiverse-Core Version | `" + this.plugin.getDescription().getVersion() + "` |" + System.lineSeparator()
                + "| Bukkit Version | `" + this.plugin.getServer().getVersion() + "` |" + System.lineSeparator()
                + "| Loaded Worlds | `" + this.plugin.getMVWorldManager().getMVWorlds() + "` |" + System.lineSeparator()
                + "| Multiverse Plugins Loaded | `" + this.plugin.getPluginCount() + "` |" + System.lineSeparator()
                + "| Economy being used | `" + plugin.getEconomist().getEconomyName() + "` |" + System.lineSeparator()
                + "| Permissions Plugin | `" + this.plugin.getMVPerms().getType() + "` |" + System.lineSeparator()
                + "## Parsed Config" + System.lineSeparator()
                + "These are what Multiverse thought the in-memory values of the config were." + System.lineSeparator() + System.lineSeparator()
                + "| Config Key  | Value |" + System.lineSeparator()
                + "| --- | --- |" + System.lineSeparator()
                + "| version | `" + this.plugin.getMVConfig().getVersion() + "` |" + System.lineSeparator()
                + "| messagecooldown | `" + plugin.getMessaging().getCooldown() + "` |" + System.lineSeparator()
                + "| teleportcooldown | `" + plugin.getMVConfig().getTeleportCooldown() + "` |" + System.lineSeparator()
                + "| worldnameprefix | `" + plugin.getMVConfig().getPrefixChat() + "` |" + System.lineSeparator()
                + "| worldnameprefixFormat | `" + plugin.getMVConfig().getPrefixChatFormat() + "` |" + System.lineSeparator()
                + "| enforceaccess | `" + plugin.getMVConfig().getEnforceAccess() + "` |" + System.lineSeparator()
                + "| displaypermerrors | `" + plugin.getMVConfig().getDisplayPermErrors() + "` |" + System.lineSeparator()
                + "| teleportintercept | `" + plugin.getMVConfig().getTeleportIntercept() + "` |" + System.lineSeparator()
                + "| firstspawnoverride | `" + plugin.getMVConfig().getFirstSpawnOverride() + "` |" + System.lineSeparator()
                + "| firstspawnworld | `" + plugin.getMVConfig().getFirstSpawnWorld() + "` |" + System.lineSeparator()
                + "| debug | `" + plugin.getMVConfig().getGlobalDebug() + "` |" + System.lineSeparator();
    }

    private void addVersionInfoToEvent(MVVersionEvent event) {
        // add the legacy version info
        event.appendVersionInfo(this.getLegacyString());

        // add the legacy file, but as markdown so it's readable
        event.putDetailedVersionInfo("version.md", this.getMarkdownString());

        // add config.yml
        File configFile = new File(this.plugin.getDataFolder(), "config.yml");
        event.putDetailedVersionInfo("version.md", configFile);

        // add worlds.yml
        File worldsFile = new File(this.plugin.getDataFolder(), "worlds.yml");
        event.putDetailedVersionInfo("version.md", worldsFile);
    }

    @Override
    public void runCommand(final CommandSender sender, final List<String> args) {
        // Check if the command was sent from a Player.
        if (sender instanceof Player) {
            sender.sendMessage("Version info dumped to console. Please check your server logs.");
        }

        MVVersionEvent versionEvent = new MVVersionEvent();

        this.addVersionInfoToEvent(versionEvent);
        this.plugin.getServer().getPluginManager().callEvent(versionEvent);

        if (CommandHandler.hasFlag("--include-plugin-list", args)) {
            versionEvent.appendVersionInfo(System.lineSeparator() + "Plugins: " + getPluginList());
            versionEvent.putDetailedVersionInfo("plugins.txt", "Plugins: " + getPluginList());
        }

        final String versionInfo = versionEvent.getVersionInfo();
        final Map<String, String> files = versionEvent.getDetailedVersionInfo();

        // log to console
        String[] lines = versionInfo.split(System.lineSeparator());
        for (String line : lines) {
            if (!line.isEmpty()) {
                Logging.info(line);
            }
        }

        BukkitRunnable logPoster = new BukkitRunnable() {
            @Override
            public void run() {
                if (args.size() > 0) {
                    String pasteUrl;
                    if (CommandHandler.hasFlag("-b", args)) {
                        // private post to pastebin
                        pasteUrl = postToService(PasteServiceType.PASTEBIN, true, versionInfo, files);
                    } else if (CommandHandler.hasFlag("-g", args)) {
                        // private post to github
                        pasteUrl = postToService(PasteServiceType.GITHUB, true, versionInfo, files);
                    } else if (CommandHandler.hasFlag("-h", args)) {
                        // private post to hastebin
                        pasteUrl = postToService(PasteServiceType.HASTEBIN, true, versionInfo, files);
                    } else if (CommandHandler.hasFlag("-p", args)) {
                        // private post to paste.gg
                        pasteUrl = postToService(PasteServiceType.PASTEGG, true, versionInfo, files);
                    } else {
                        return;
                    }

                    if (!(sender instanceof ConsoleCommandSender)) {
                        sender.sendMessage("Version info dumped here: " + ChatColor.GREEN + pasteUrl);
                    }
                    Logging.info("Version info dumped here: %s", pasteUrl);
                }
            }
        };

        // Run the log posting operation asynchronously, since we don't know how long it will take.
        logPoster.runTaskAsynchronously(this.plugin);
    }

    /**
     * Send the current contents of this.pasteBinBuffer to a web service.
     *
     * @param type       Service type to send paste data to.
     * @param isPrivate  Should the paste be marked as private.
     * @param pasteData  Legacy string only data to post to a service.
     * @param pasteFiles Map of filenames/contents of debug info.
     * @return URL of visible paste
     */
    private static String postToService(PasteServiceType type, boolean isPrivate, String pasteData, Map<String, String> pasteFiles) {
        PasteService ps = PasteServiceFactory.getService(type, isPrivate);

        try {
            String result;
            if (ps.supportsMultiFile()) {
                result = ps.postData(pasteFiles);
            } else {
                result = ps.postData(pasteData);
            }

            if (SHORTENER != null) return SHORTENER.shorten(result);
            return result;
        } catch (PasteFailedException e) {
            e.printStackTrace();
            return "Error posting to service.";
        } catch (NullPointerException e) {
            e.printStackTrace();
            return "That service isn't supported yet.";
        }
    }

    private String getPluginList() {
        return StringUtils.join(plugin.getServer().getPluginManager().getPlugins(), ", ");
    }
}
